/*
  dmp.h

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include <cstdint>
#include <memory>
#include <vector>
#include "acn/pdu.h"

// Architecture for Control Networks – Device Management Protocol
namespace ACN::DMP {

/**
 * @brief The data_type_t enum
 *
 * @cite DMP 5.1.4 Address and Data Types
 * D1, D0 = Specify non-range or range address, single data, equal size or
 * mixed size data array
 */
enum data_type_t {
  SINGLE = 0b00, //!< Non-range address, Single data item
  RANGE  = 0b01, //!< Range address, Single data item
  ARRAY  = 0b10, //!< Range address, Array of equal size data items
  SERIES = 0b11  //!< Range address, Series of mixed size data items
};


/**
 * @brief The address_element_length enum
 */
enum element_length {
  ONE  = 0b00, // 0
  TWO  = 0b01, // 1
  FOUR = 0b10, // 2
  ZERO = 0b11  // 3 (reserved)
};


/**
 * @brief The address_type struct
 */
struct address_type
    : PDU::pdu_header
{
  element_length address_length;  //!< A1, A0 = Size of Address elements
  uint8_t        x_reserved;      //!< X1, X0
  data_type_t    data_type;       //!< D1, D0
  bool           relative;        //!< R = Specifies whether address is relative or not.
  bool           z_reserved;      //!< Z

  size_t streamSize() const override { return 1; }
  void   iStream(PDU::Stream) override;
  void   oStream(PDU::Stream) const override;
};


/**
 * @brief 5.1.5 The range struct
 */
/// \todo maybe template this struct based on address_length
struct range
    : PDU::pdu_stream_object
{
  /**
   * @brief range
   * @param t
   */
  range(const address_type& t) : type_(t) {};

  uint32_t address   = 0;   //!< start address
  uint32_t incriment = 0;   //!< property size (number of octets)
  uint32_t count     = 0;   //!< number of properties

  size_t streamSize() const override;
  void   iStream(PDU::Stream) override;
  void   oStream(PDU::Stream) const override;

private:
  const address_type& type_;

  uint32_t read_(PDU::Stream);
  void     write_(PDU::Stream, const uint32_t&) const;
};


/**
 * @brief set_property
 */
typedef std::pair<range, std::vector<uint8_t>> address_data_pair;


/**
 * @brief The address_pair_list struct
 */
struct address_pair_list
    : PDU::pdu_data
{
  /**
   * @brief address_pair_list
   * @param t
   */
  address_pair_list(const address_type& t) : type_(t) {};
  std::vector<address_data_pair> properties;  //!< list of address/data pairs

  size_t streamSize() const override;
  void   iStream(PDU::Stream) override;
  void   oStream(PDU::Stream) const override;

private:
  const address_type& type_;
};


/**
 * @brief The address_list struct
 */
struct address_list
    : PDU::pdu_data
{
  /**
   * @brief address_list
   * @param t
   */
  address_list(const address_type& t) : type_(t) {};
  std::vector<range> addresses; //!< list of addresses

  size_t streamSize() const override;
  void   iStream(PDU::Stream) override;
  void   oStream(PDU::Stream) const override;

private:
  const address_type& type_;
};


/**
 * @brief 7 Response Messages
 */
enum failure_reason {
  NONSPECIFIC  = 1, //!< Non-specific or non-DMP reason.
  NOT_PROPERTY = 2, //!< The address does not correspond to a property.
  WRITE_ONLY   = 3, //!< The property’s value may not be read.
  NOT_WRITABLE = 4, //!< The property’s value may not be written.
  DATA_ERROR   = 5, //!< The data does not correspond to the property.
  SUBSCIRPTION_NOT_SUPPORTED = 10, //!< Subscriptions on the specified property are not supported by the device.
  NO_SUBSCRIPTIONS_SUPPORTED = 11, //!< Subscriptions not supported on any property.
  INSUFFICIENT_RESOURCES = 12, //!< The component cannot support more subscriptions due to resource limitations
  UNAVAILABLE  = 13 //!< The property’s value is not available due to restrictions imposed by device specific functionality (e.g., access permission mechanisms).
};


/**
 * @brief 13.1 Protocol Codes
 */
static const uint32_t DMP_PROTOCOL_ID   = 2; //!< protocol vector

// 13.2 Message Codes
static const uint8_t GET_PROPERTY       = 1;
static const uint8_t SET_PROPERTY       = 2;
static const uint8_t GET_PROPERTY_REPLY = 3;
static const uint8_t EVENT              = 4;
static const uint8_t SUBSCRIBE          = 7;
static const uint8_t UNSUBSCRIBE        = 8;
static const uint8_t GET_PROPERTY_FAIL  = 9;
static const uint8_t SET_PROPERTY_FAIL  = 10;
static const uint8_t SUBSCRIBE_ACCEPT   = 12;
static const uint8_t SUBSCRIBE_REJECT   = 13;
static const uint8_t SYNC_EVENT         = 17;


/**
 * @brief The DMP::Pdu class
 */
class Pdu
  : public PDU::Pdu
{
public:
  Pdu();
  void iStream(PDU::Stream) override;
};


} // ACN::DMP
