/*
  mergeproxyuniverse.h

  Copyright (c) 2021 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "universe.h"

#include <memory>
#include <unordered_map>

namespace sACN {

/**
 * @brief Priority based selection of multiple source universes
 *
 *
 * > \cite sACN 6.2.3.1 Multiple Sources at Highest Priority
 * >
 * > It is possible for there to be multiple sources, all transmitting data
 * > at the highest currently active priority for a given universe. When this
 * > occurs, receivers must handle these sources in some way.
 *
 * If multiple universes with highest priority are recieved, the universe
 * encountered first will maintain dominance.
 *
 * > \cite sACN 6.2.3.2 Note on Merge and Arbitration Algorithms
 * >
 * > A process which selects between candidate sources based on some
 * > additional selection criterion is called arbitration.
 *
 * The universe with the highest priority will be determined to be the
 * dominant source.  Universes with NULL Start Code frequency less than
 * DMX::E111_DATA_LOSS_TIMEOUT are not considered as eligible source universes.
 */
class ArbitratingUniverse
  : public sACN::Universe
{
public:
    ArbitratingUniverse();
    ~ArbitratingUniverse();

    uint16_t expectedUniverse; ///< Expected universe number

    void setHoldLastLook(const bool);
    bool getHoldLastLook() const;

    // Source universes:
    void deleteSourceUniverse(const DATA::data_header&);
    void dataChangedNotifier(DMX::Universe* universe);

    // SACN::Universe overrides:
    void set(std::shared_ptr<ACN::DMP::Pdu>, std::shared_ptr<DATA::data_header>) override;

    std::shared_ptr<DATA::data_header> provenance() override;
    void setProvenance(std::shared_ptr<DATA::data_header>) override {};

    bool isSyncronized() override;
    void synchronize(uint8_t = 0) override;

    bool isEditable() const override;

    uint16_t activeSlots() override;

    // api for poly-source universes
    const std::vector<DATA::data_header> sources() const override;
    std::shared_ptr<Universe> sourceUniverse(const DATA::data_header&) override;
    std::shared_ptr<Universe> addNewSource(const DATA::data_header&) override;
    std::shared_ptr<void> onSourceListChange(std::function<void()>) override;

    // DMX::Universe Overrides:
    uint8_t slot(const uint16_t) override;
    double rxRate() override;

protected:
  void doListChangeCallbacks();

private:
  std::unordered_map<DATA::data_header, std::shared_ptr<Universe>> sources_;
  std::vector<std::shared_ptr<void>> source_data_tokens; //!< source universe data change tokens
  std::vector<std::weak_ptr<std::function<void()>>> cb_sourceListChange;  //!< list of calback functions

  bool HoldLastLook;

  std::shared_ptr<Universe> dominant_();
  bool hasSourceUniverse(const DATA::data_header&) const;
};

} // SACN namespace
