/*
  extended.h

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "acn/pdu.h"
#include "sacn.h"
#include "uuid/uuid.h"

/**
 * @brief both Discovery and Syncronization
 */
namespace sACN::EXTENDED {

/**
 * @brief \cite sACN 6.3 E1.31 Synchronization Packet Framing Layer
 */
struct sync_header
    : ACN::PDU::pdu_header
{
  /// @brief \cite sACN 6.3.2 Synchronization Packet: Sequence Number
  ///
  /// > Sources shall maintain a sequence for each universe they transmit.
  /// > The sequence number for a universe shall be incremented by one for
  /// > every packet sent on that universe. There is no implied relationship
  /// > between the sequence number of an E1.31 Synchronization Packet and
  /// > the sequence number of an E1.31 Data Packet on that same universe.
  uint8_t sequence_number;

  /// @brief \cite sACN 6.3.3 Synchronization Packet: Synchronization Address
  ///
  /// > The Synchronization Address identifies the universe to which this
  /// > synchronization packet is directed.
  uint16_t sync_address;

  /// @brief \cite sACN 6.3.4 Synchronization Packet: Reserved
  ///
  /// > Octets 47-48 of the E1.31 Synchronization Packet are reserved for
  /// future use. They shall be transmitted as 0 and ignored by receivers.
  uint16_t reserved;

  size_t streamSize() const override { return 5; }
  void iStream(ACN::PDU::Stream) override;
  void oStream(ACN::PDU::Stream) const override;
};


/**
 * @brief \cite sACN 6.4 E1.31 Universe Discovery Packet Framing Layer
 */
struct discovery_header
    : ACN::PDU::pdu_header
{
  /// @brief \cite sACN 6.4.2 E1.31 Universe Discovery Packet: Source Name
  ///
  /// > A user-assigned name provided by the source of the packet for use in
  /// > displaying the identity of a source to a user. There is no mechanism,
  /// > other than user configuration, to ensure uniqueness of this name. The
  /// > source name shall be null-terminated. If the source component
  /// > implements ACN discovery as defined in EPI 19 \cite epi19, then this
  /// > name shall be the same as the UACN field specified in EPI 19
  /// > \cite epi19. User-Assigned Component Names, as the title suggests,
  /// > supply a single name for an entire component, so this Source Name
  /// > field will exist for each unique CID, but may be the same across
  /// > multiple universes sourced by the same component.
  std::string source_name;

  /// @brief \cite sACN 6.4.3 E1.31 Universe Discovery Packet: Reserved
  ///
  /// Octets 108-111 of the E1.31 Universe Discovery Packet are reserved for
  /// future use. They shall be transmitted as 0 and ignored by receivers.
  uint32_t reserved;

  size_t streamSize() const override { return 68; }
  void iStream(ACN::PDU::Stream) override;
  void oStream(ACN::PDU::Stream) const override;
};


/**
 * @brief The EXTENDED::Pdu class
 */
class Pdu
    : public ACN::PDU::Pdu
{
public:
  Pdu();
  void iStream(ACN::PDU::Stream) override;
};


/**
 * @brief \cite sACN 1.6 Universe Discovery
 *
 * > This standard includes a Universe Discovery packet that sources must
 * > provide in order to enumerate the universes upon which they are
 * > transmitting. This allows other devices interested in network traffic to
 * > monitor which universes are currently active, without the need to join
 * > every multicast group to examine their individual transmissions.
 */
namespace DISCOVERY {

/**
 * @brief Table 8-9: E1.31 Universe Discovery Packet Universe Discovery Layer
 */
struct discovery_list_header
    : ACN::PDU::pdu_header
{
  uint8_t page;         //!< number of the current page
  uint8_t last_page;    //!< total pages

  size_t streamSize() const override { return 2; }
  void iStream(ACN::PDU::Stream) override;
  void oStream(ACN::PDU::Stream) const override;
};


/**
 * @brief The discoveredUniverse struct
 */
struct discoveredUniverse
{
  UUID::uuid  source;       //!< CID of source
  std::string description;  //!< universe description
  uint16_t    universe;     //!< universe number

  friend bool operator== (const discoveredUniverse& a,
                          const discoveredUniverse& b);
  friend bool operator< (const discoveredUniverse& a,
                         const discoveredUniverse& b);
};


/**
 * @brief The discovery_list_data struct
 */
struct discovery_list_data
    : ACN::PDU::pdu_data
{
  std::vector<std::shared_ptr<discoveredUniverse>> found; //!< discovered universes

  size_t streamSize() const override { return 2 * found.size(); }
  void iStream(ACN::PDU::Stream) override;
  void oStream(ACN::PDU::Stream) const override;
};


/**
 * @brief \cite sACN 3.10 E1.31 Universe Discovery Packet
 *
 * > An E1.31 Universe Discovery Packet is a packet which contains a packed
 * > list of the universes upon which a source is actively operating. It is
 * > transmitted with the VECTOR_E131_EXTENDED_DISCOVERY vector.
 */
class Pdu
    : public ACN::PDU::Pdu
{
public:
  Pdu();
  void iStream(ACN::PDU::Stream) override;
};


} // DISCOVERY
} // SACN::EXTENDED

namespace std
{
  /**
   * @brief The hash struct specialization for
   * sACN::EXTENDED::DISCOVERY::discoveredUniverse
   */
  template<>
  struct hash<sACN::EXTENDED::DISCOVERY::discoveredUniverse>
  {
    /**
     * @brief operator ()
     * @param disc
     * @return
     */
    size_t operator()(sACN::EXTENDED::DISCOVERY::discoveredUniverse const& disc) const noexcept
    {
      size_t h1 = hash<string>{}(disc.description);
      size_t h2 = hash<uint16_t>{}(disc.universe);
      size_t h3 = hash<UUID::uuid>{}(disc.source);
      size_t h = h1 ^ h2 ^ h3; // or use boost::hash_combine
      return h;
    }
  };
} // namespace std
