/*
  sdt/udp.h

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "acn/pdu.h"
#include <cstdint>

// ACN EPI 18 - Operation of SDT on UDP Networks
namespace ACN::SDT::UDP {

//  Table 3. SDT symbolic parameters
static const float    MAK_TIMEOUT_FACTOR   = 0.1;
static const uint8_t  MAK_MAX_RETRIES      = 2;    // (3 tries total)
static const uint8_t  AD_HOC_TIMEOUT       = 200;  // ms
static const uint8_t  AD_HOC_RETRIES       = 2;    // (3 tries total)
static const float    RECIPROCAL_TIMEOUT_FACTOR = 0.2;
static const uint8_t  MIN_EXPIRY_TIME      = 2;    // s
static const float    NAK_TIMEOUT_FACTOR   = 0.1;
static const uint8_t  NAK_MAX_RETRIES      = 2;    // (3 tries total)
static const uint8_t  NAK_HOLDOFF_INTERVAL = 2;    // ms
static const uint8_t  NAK_MAX_TIME         = 10 * NAK_HOLDOFF_INTERVAL;
static const uint8_t  NAK_BLANKTIME        = 3  * NAK_HOLDOFF_INTERVAL;
static const uint16_t SDT_MULTICAST_PORT   = 5568; // IANA registered “sdt”


/**
 * @brief 3 Address Specification
 */
struct ipAddress
  : PDU::pdu_stream_object
{
  ipAddress()
    : type(0) // SDT_ADDR_NULL
    , port(SDT_MULTICAST_PORT)
  {
    memset(address.ipv6.bytes, 0, sizeof(address.ipv6.bytes));
    address.ipv6.scope = 0x8;
    address.ipv6.flag_t = true;
    address.ipv6.prefix = 0xff;
  };

  uint8_t  type; //!< address type identifier
  uint16_t port; //!< TCP/UDP Port
  union {
    struct {
      union {
        uint8_t  bytes[4];
        uint32_t value;
      };
    } ipv4;
    struct {
      union {
        uint8_t bytes[16];
        struct {
          uint8_t group[14];
          union {
            uint8_t flags_and_scope;
            struct {
              /// > The scope shall be set to 0x8 to indicate Organization-Local
              /// > scope per RFC 4291 \cite IPv6Addr and RFC 7346 \cite IPv6McastScope.
              uint8_t scope  : 4;
              /// > The T flag indicates a “transient” multicast address that
              /// > is not permanently assigned by IANA. This bit shall be set
              /// > to 1 to indicate a non-permanently assigned multicast address.
              bool flag_t : 1;
              /// > The P flag is defined by RFC 3306 \cite UnicastPrefixMcast
              /// > and shall be set to 0.
              bool flag_p : 1;
              /// > The R flag is defined by RFC 3956 \cite EmbeddedRP and
              /// > shall be set to 0.
              bool flag_r : 1;
              /// > The high-order flag bit is reserved and shall be set to 0.
              bool flag_0 : 1;
            };
          };
          /// > The multicast address prefix is fixed at 0xFF.
          uint8_t prefix;
        };
      };
    } ipv6;
  } address; ///< ipv4 or ipv6 address

  size_t streamSize() const override;
  void iStream(PDU::Stream) override;
  void oStream(PDU::Stream) const override;
};

} // ACN::SDT::UDP
